unit UnitMain;

{$MODE Delphi}

interface

uses
  Windows, Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls,
  ExtCtrls, Buttons, Grids, Menus, ChessBrd;

const engine_filename: string = 'stockfish_32bit.exe';

type
  TMyThread = class(TThread)
  private
    TextString: string;
    procedure UpdateCaption;
  protected
    procedure Execute; override;
  end;

type

  { TForm1 }

  TForm1 = class(TForm)
    backbutton: TButton;
    forwardbutton: TButton;
    Button1: TButton;
    Button4: TButton;
    Edit1: TEdit;
    MainMenu1: TMainMenu;
    MenuItem1: TMenuItem;
    MenuItem2: TMenuItem;
    MenuItem3: TMenuItem;
    MenuItem4: TMenuItem;
    MenuItem5: TMenuItem;
    ML: TStringGrid;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    procedure FormActivate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure backbuttonClick(Sender: TObject);
    procedure forwardbuttonClick(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure MenuItem2Click(Sender: TObject);
    procedure MenuItem3Click(Sender: TObject);
    procedure MenuItem5Click(Sender: TObject);
    procedure MLDrawCell(Sender: TObject; aCol, aRow: Integer; aRect: TRect; aState: TGridDrawState);
    procedure MLKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure MLMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  private
    { private declarations }
    procedure CreateChessBoard;
    procedure Update_MoveList;
    procedure Update_Movelist_Selection;
    procedure Load_Engine;
    procedure ChessBoardLegalMove(Sender: TObject; oldSq, newSq: Square);
  public
    { public declarations }
  end; 

var
  Form1: TForm1;
  CB: TChessBrd;
  StartInfo: TStartupInfo;
  ProcInfo: TProcessInformation;
  Security: TSecurityAttributes;
  InputRead, InputWrite, OutputRead, OutputWrite, ErrorRead, ErrorWrite: THandle;
  myThread: TThread;
  Created: boolean = False;
  Board_Created: boolean = False;

implementation

{$R *.lfm}

function ReadPipeInput(InputPipe: THandle; var BytesRem: Integer): string;
var
 TextBuffer: array[1..32767] of char;
 TextString: string;
 BytesRead: longword;
 PipeSize: Integer;
begin
 Result:='';
 PipeSize:=Sizeof(TextBuffer);
 PeekNamedPipe(InputPipe, nil, PipeSize, @BytesRead, @PipeSize, @BytesRem);
 if bytesread>0 then
 begin
  ReadFile(InputPipe, TextBuffer, pipesize, bytesread, nil);
  OemToChar(@TextBuffer, @TextBuffer);
  TextString:=string(TextBuffer);
  SetLength(TextString, BytesRead);
  Result:=TextString;
 end;
end;

procedure WritePipeOut(OutputPipe: THandle; InString: string);
var BytesWritten: Dword;
begin
 InString:=InString+#13#10;
 WriteFile(OutputPipe, Instring[1], Length(Instring), BytesWritten, nil);
end;

procedure Send_Position_To_Engine;
begin
 WritePipeOut(InputWrite, 'position fen '+CB.GetFenPosition);
end;

procedure Start_Thinking;
begin
 WritePipeOut(InputWrite, 'go movetime 3000');
end;

procedure TMyThread.Execute;
var BytesRem: longint;
begin
 while not Terminated do
 begin
  TextString:=ReadPipeInput(OutputRead, BytesRem);
  if TextString<>'' then Synchronize(UpdateCaption);
  TextString:=ReadPipeInput(ErrorRead, BytesRem);
  if TextString<>'' then Synchronize(UpdateCaption);
  sleep(100);
 end;
end;

procedure TMyThread.UpdateCaption;
var m, c, tmp: string;
p: integer;
begin
 tmp:='';
 p:=pos('bestmove', TextString);
 if p>0 then
 begin
  m:=UPPERCASE(copy(TextString, p+9, 4));
  c:=copy(TextString, p+13, 1);
  if (c<>' ') and (length(c)=1) then CB.PromoteTo:=c[1];
  CB.Performmove(CB.stringtosquare(copy(m,1,2)), CB.stringtosquare(copy(m,3,2)));
 end;
 p:=pos('ponder', TextString);
 if p>0 then
 begin
  tmp:='Hint: '+copy(TextString, p+7, 4);
 end;
 form1.edit1.text:=tmp;
end;

procedure TForm1.Load_Engine;
var fn: string;
begin
 fn:=StringReplace(ExtractFilePath(ParamStr(0)), '/', '\', [rfreplaceall])+engine_filename;
 if not fileexists(fn) then
 begin
  showmessage('File not found'#13#10#13#10+fn);
  exit;
 end;
 if Created=true then exit;
 Created:=true;
 with Security do
 begin
  nlength:=SizeOf(TSecurityAttributes);
  binherithandle:=true;
  lpsecuritydescriptor:=nil;
 end;
 CreatePipe(InputRead, InputWrite, @Security, 0);
 CreatePipe(OutputRead, OutputWrite, @Security, 0);
 CreatePipe(ErrorRead, ErrorWrite, @Security, 0);
 FillChar(StartInfo, SizeOf(TStartupInfo), #0);
 StartInfo.cb:=SizeOf(TStartupInfo);
 StartInfo.hStdInput:=InputRead;
 StartInfo.hStdOutput:=OutputWrite;
 StartInfo.hStdError:=ErrorWrite;
 StartInfo.dwFlags:=STARTF_USESTDHANDLES+STARTF_USESHOWWINDOW;
 StartInfo.wShowWindow:=SW_HIDE;
 if CreateProcess(nil, PChar(fn), @Security, @Security, true,
    CREATE_NEW_CONSOLE+SYNCHRONIZE, nil, nil, StartInfo, ProcInfo) then
 begin
  MyThread:=TMyThread.Create(false);
  MyThread.Priority:=tpHigher;
 end else showmessage('Failed!');
end;

function copy_to_end(source: string; start: integer): string;
begin
 Result:=copy(source, start, length(source)-start+1);
end;

procedure Clear_Movelist;
var i: integer;
begin
 for i:=0 to form1.ML.rowcount-1 do
  form1.ML.rows[i].Clear;
end;

procedure TForm1.Update_MoveList;
var
 MoveList: TStringList;
 tmp: string;
 p, i, count: integer;
begin
 Clear_Movelist;
 MoveList:=TStringList.Create;
 CB.GetMoveList(MoveList);
 count:=MoveList.count;
 if count>0 then ML.rowcount:=count else ML.rowcount:=1;
 for i:=0 to count-1 do
 begin
  tmp:=movelist.strings[i];
  p:=pos('.', tmp);
  ML.Cells[0,i]:=copy(tmp, 1, p);
  tmp:=copy_to_end(tmp, p+1);
  p:=pos(',', tmp);
  if p>0 then
  begin
   ML.Cells[1,i]:=copy(tmp, 1, p-1);
   ML.Cells[2,i]:=copy_to_end(tmp, p+1);
   ML.col:=2;
  end else
  begin
   ML.Cells[1,i]:=tmp;
   ML.Cells[2,i]:='';
   ML.col:=1;
  end;
 end;
 ML.SetFocus;
 ML.Row:=ML.RowCount-1;
 MoveList.Free;
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
 CB.ComputerPlaysWhite:=false;
 CB.newgame;
 CB.ComputerPlaysBlack:=true;
 WritePipeOut(InputWrite, 'ucinewgame');
 Update_MoveList;
 if (CB.WhiteToMove) and (CB.ComputerPlaysWhite) then Start_Thinking;
end;

procedure TForm1.ChessBoardLegalMove(Sender:TObject; oldSq, newSq: Square);
begin
 Update_MoveList;
 Send_Position_To_Engine;
 if (((not CB.WhiteToMove) and (CB.ComputerPlaysBlack)) or ((CB.WhiteToMove) and (CB.ComputerPlaysWhite))) then
   Start_Thinking;
end;

procedure TForm1.CreateChessBoard;
begin
 CB:=TChessBrd.Create(Form1);
 CB.Left:=0;
 CB.Top:=0;
 CB.Parent:=Form1;
 CB.Language:=cblFinnish;
 //CB.Notation:=cblcomplete;
 CB.Animatemoves:=false;
 CB.CustomEngine:=true;
 CB.ComputerPlaysBlack:=true;
 CB.PromotionPiece:=cblAsk;
 CB.BorderColor:=clGray;
 CB.SquareColorDark:=clGray;
 CB.SquareColorLight:=clSilver;
 CB.OnLegalMove:=ChessBoardLegalMove;
 CB.DisplayCoords:=[North, East, West, South];
end;

procedure TForm1.MLKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
 if key=vk_left then backbuttonclick(self);
 if key=vk_right then forwardbuttonclick(self);
 if (key=vk_up) or (key=vk_down) then
   if ML.col=2 then CB.GotoMove(ML.Row+2, true) else CB.GotoMove(ML.Row+1, false);
end;

procedure TForm1.MLMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 if ML.col=2 then CB.GotoMove(ML.Row+2, true) else CB.GotoMove(ML.Row+1, false);
end;

procedure TForm1.Update_Movelist_Selection;
var x, y: integer;
begin
 if CB.WhiteToMove then
 begin
  x:=2;
  y:=CB.currentmove-2;
 end
 else
 begin
  x:=1;
  y:=CB.currentmove-1;
 end;
 if y<0 then begin y:=0; x:=0; end;
 ML.col:=x;
 ML.row:=y;
end;

procedure TForm1.backbuttonClick(Sender: TObject);
begin
 CB.MoveBackward;
 Update_MoveList;
 Update_Movelist_Selection;
end;

procedure TForm1.FormActivate(Sender: TObject);
begin
 if Board_Created=false then
 begin
  CreateChessBoard;
  Board_Created:=true;
  Load_Engine;
 end;
end;

procedure TForm1.forwardbuttonClick(Sender: TObject);
begin
 CB.MoveForward;
 Update_MoveList;
 Update_Movelist_Selection;
end;

procedure TForm1.Button4Click(Sender: TObject);
begin
 if CB.ComputerplaysWhite then
 begin
  CB.ComputerPlaysWhite:=false;
  CB.ComputerPlaysBlack:=true;
  if (not CB.WhiteToMove) then
  begin
   Send_Position_To_Engine;
   Start_Thinking;
  end;
 end else
 if CB.ComputerplaysBlack then
 begin
  CB.ComputerPlaysBlack:=false;
  CB.ComputerPlaysWhite:=true;
  if (CB.WhiteToMove) then
  begin
   Send_Position_To_Engine;
   Start_Thinking;
  end;
 end;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
 if Created=true then
 begin
  WritePipeOut(InputWrite, 'quit');
  MyThread.Terminate;
  MyThread.free;
  CloseHandle(ProcInfo.hProcess);
  CloseHandle(ProcInfo.hThread);
  CloseHandle(InputRead);
  CloseHandle(InputWrite);
  CloseHandle(OutputRead);
  CloseHandle(OutputWrite);
  CloseHandle(ErrorRead);
  CloseHandle(ErrorWrite);
 end;
end;

procedure TForm1.MenuItem2Click(Sender: TObject);
begin
  cb.computerplayswhite:=false;
  cb.computerplaysblack:=false;
  if opendialog1.execute then cb.LoadPgn(opendialog1.filename, 1);
end;

procedure TForm1.MenuItem3Click(Sender: TObject);
begin
  if savedialog1.execute then
  begin
   if fileexists(savedialog1.filename) then
    if messageDlg(Format('File exists!'+#13#13+ 'Overwrite %s', [savedialog1.filename]), mtConfirmation, mbYesNoCancel, 0)<>idYes then exit;
   cb.SavePgn(savedialog1.filename);
  end;
end;

procedure TForm1.MenuItem5Click(Sender: TObject);
begin
  close;
end;

procedure TForm1.MLDrawCell(Sender: TObject; aCol, aRow: Integer; aRect: TRect; aState: TGridDrawState);
begin
 if (aRow=ML.Row) and (aCol=ML.Col) and (aCol>0) then
 begin
  ML.Canvas.Brush.Color:=Clskyblue;
  ML.Canvas.FillRect(aRect);
  ML.Canvas.Textout(aRect.Left+2, aRect.Top+2, ML.Cells[aCol, aRow])
 end;
end;

end.
